/*
 * bmchealth.c
 *
 * This tool checks the health of the BMC via IPMI.
 * 
 * Author:  Andy Cress  arcress@users.sourceforge.net
 * Copyright (c) 2006 Intel Corporation. 
 *
 * 03/22/06 Andy Cress - created
 * 06/20/06 Andy Cress 0.6 - more vendor strings, add ping_node() stub for now
 * 10/20/06 Andy Cress 1.1 - added -g for guid
 */
/*M*
Copyright (c) 2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#endif
#include <string.h>
#include "ipmicmd.h"
 
#define  SELFTEST_STATUS   0x04
/*
 * Global variables 
 */
char * progname  = "bmchealth";
char * progver   = "1.2";
char   fdebug    = 0;
char   fpicmg    = 0;
char   do_hsc    = 0;
char   do_guid   = 0;

static int get_selftest_status(uchar *rdata, int rlen)
{
	uchar idata[4];
	uchar ccode;
	int ret;

        ret = ipmi_cmdraw( SELFTEST_STATUS, NETFN_APP,
                        BMC_SA,PUBLIC_BUS,BMC_LUN,
                        idata,0, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;
	return(ret);
}  /*end get_selftest_status()*/

static int get_hsc_devid(uchar *rdata, int rlen)
{
	uchar ccode;
	int ret;

        ret = ipmi_cmdraw( 0x01, /*(GET_DEVICEID & 0x00ff)*/ 
			NETFN_APP, 0xC0,PUBLIC_BUS,BMC_LUN,
                        NULL,0, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;
	return(ret);
}  /*end get_hsc_devid*/

static int get_chassis_status(uchar *rdata, int rlen)
{
	uchar idata[4];
	uchar ccode;
	int ret;

        ret = ipmi_cmdraw( 0x01, NETFN_CHAS, BMC_SA, PUBLIC_BUS,BMC_LUN,
                        idata,0, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;
	return(ret);
}  /*end get_chassis_status*/

static void show_chassis_status(uchar *data, int len)
{
    uchar fault;
    char faultstr[12];
    char *state;

    if (data[0] & 0x01) state = "On";
    else state = "OFF";
    fault = (data[0] & 0x1E);
    if (fault != 0) sprintf(faultstr,"Fault: %02x",fault);
    else strcpy(faultstr,"OK");
    printf("Chassis Status   = %02x %s, %s\n",data[0],state,faultstr);
}

static int get_chan_auth(uchar *rdata, int rlen)
{
	uchar idata[4];
	uchar ccode;
	int ret;

        idata[0] = 0x0e; /*this channel*/
        idata[1] = 0x02; /*user level*/
        ret = ipmi_cmdraw( 0x38, /*(GET_CHAN_AUTH & 0x00ff)*/ 
			NETFN_APP, BMC_SA, PUBLIC_BUS,BMC_LUN,
                        idata,2, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;
	return(ret);
}  /*end get_chan_auth*/

void show_chan_auth(char *tag, uchar *rec, int srec)
{
   char pstr[40];
   pstr[0] = 0;
   if (rec[1] & 0x01) strcat(pstr,"None ");
   if (rec[1] & 0x02) strcat(pstr,"MD2 ");
   if (rec[1] & 0x04) strcat(pstr,"MD5 ");
   if (rec[1] & 0x10) strcat(pstr,"Straight_Passwd ");
   if (rec[1] & 0x20) strcat(pstr,"OEM ");
   printf("Channel %d Auth Types: %s\n",rec[0],pstr);
   printf("          Status = %02x, OEM ID %02x%02x%02x OEM Aux %02x\n",
          rec[2],rec[4],rec[5],rec[6],rec[7]);
}

#define BMC  1
#define HSC  2

void show_devid(int dtype, uchar *devrec, int sdevrec)
{
   uchar ipmi_maj = 0;
   uchar ipmi_min = 0;
   char *tag; 
   int mfg, prod;
   char *mfgstr = "";
   char *prodstr = "";

      ipmi_maj = devrec[4] & 0x0f;
      ipmi_min = devrec[4] >> 4;
      mfg  = devrec[6] + (devrec[7] << 8) + (devrec[8] << 16);
      prod = devrec[9] + (devrec[10] << 8);
      if (dtype == HSC) tag = "HSC";
      else tag = "BMC";
      /* For a list of all IANA enterprise mfg vendor numbers, 
       * see http://www.iana.org/assignments/enterprise-numbers 
       * Product numbers are different for each mfg vendor. 
       */
      switch(mfg) {
         case VENDOR_NSC: mfgstr = "(NSC)";    /*=0x000322*/
		        if (dtype == BMC) tag="mBMC";  break;
         case 0x000002: mfgstr = "(IBM)";   break;
         case 0x000011: mfgstr = "(HP)";    break;
         case 0x00002A: mfgstr = "(Sun)";   break;
         case 0x000077: mfgstr = "(NEC)";   break;
         case 0x002880: mfgstr = "(Fujitsu-Siemens)"; break;
         case 0x002A7C: mfgstr = "(SuperMicro)";   break;
         case 0x000175: mfgstr = "(Tatung)";  break;
         case 0x0019FD: mfgstr = "(Tyan)";    break;
         case 0x0051EE: mfgstr = "(AMI)";     break;
         case 0x0002A2: mfgstr = "(Dell)";    break;
         case 0x00400A: mfgstr = "(Pigeon)";  break;
         case VENDOR_LMC:  mfgstr = "(LMC)";  break; /*=0x0878 with SuperMicro*/
         case VENDOR_INTEL: mfgstr = "(Intel)";    /*=0x000157*/
             switch(prod) {
                 /* TODO: add some known product tags here */
                 default:       prodstr = ""; break;
             }
             break;
         default:       mfgstr = ""; break;
      }
      printf("%s version %x.%x, IPMI version %d.%d \n",
             tag, devrec[2],  devrec[3], ipmi_maj, ipmi_min);
      printf("%s Revision     = %02x\n",tag,devrec[1]);
      printf("%s SlaveAddr    = %02x\n",tag,devrec[0]);
      printf("%s manufacturer = %06x %s, product = %04x%s\n",
		tag, mfg,mfgstr,prod,prodstr);
      /* could show product rev, if available (sdevrec >= 14) */
      if (sdevrec >= 13) {
         printf("Firmware Rev    = %02x %02x\n",devrec[11],devrec[12]);
      }
      return;
}


#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
{
   int ret = 0;
   char c;
   uchar freadonly = 1;
   uchar freset = 0;
   uchar fdisable = 0;
   uchar selfbuf[16];
   uchar devrec[16];
   uchar cbuf[16];
   uchar cc;
   int t = 0;
   int selfstatus;
   char *selfstr = "";
   char *s;
   int i, sresp;

   printf("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"ghT:V:EP:N:R:U:x?")) != EOF ) 
      switch(c) {
          case 'x': fdebug = 1;     break;  /* debug messages */
          case 'h': do_hsc = 1;     break;  /* check the HSC too */
          case 'g': do_guid = 1;    break;  /* get the System GUID also */
          case 'T':      /* auth type */
		i = atoi(optarg);  
                if (i > 0 && i <= 5) gauth_type = i;
                break;
          case 'V':      /* priv level */
		i = atoi(optarg);
                if (i > 0 && i <= 5) gpriv_level = i;
                break;
	  case 'N':      /* nodename */
		gnode = optarg;
		fipmi_lan = 1; 
		break;
	  case 'U':
		guser = strdup_(optarg);  /*remote username */
		if (!guser)  perror("strdup");
		else {  /* Hide username from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		}
		break;
 	  case 'R': 
 	  case 'P': 
		gpswd = strdup_(optarg);  /*remote password */
		if (!gpswd) perror("strdup");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		} 
 		break;
          case 'E':      /* get password from IPMI_PASSWORD environment var */
                gpswd = getenv("IPMI_PASSWORD");
                if (gpswd == NULL) perror("getenv(IPMI_PASSWORD)");
                else {
                   if (strlen(gpswd) > 16) gpswd[16] = 0;
                   if (fdebug) printf("using IPMI_PASSWORD\n");
                }
                break;
	  default:
                printf("Usage: %s [-ghx -N node -U user -P/-R pswd -ETV]\n", progname);
                printf(" where -x      show eXtra debug messages\n");
                printf("       -h      check the HotSwap Controller also\n");
                printf("       -g      get the System GUID also\n");
          printf("       -N node  Nodename of target system\n");
          printf("       -U user  Username for remote node\n");
          printf("       -P/-R pswd  Remote Password\n");
          printf("       -E   use password from Environment IPMI_PASSWORD\n");
          printf("       -T 1 use auth Type: 1=MD2, 2=MD5(default)\n");
          printf("       -V 2 use priVilege level: 2=user, 4=admin(default)\n");
                exit(1);
      }

   ret = ipmi_getdeviceid(devrec,16,fdebug);
   if (ret != 0) {
	printf("Cannot do ipmi_getdeviceid, ret = %d\n",ret);
	ipmi_close_();
	exit(1);
   } else {
       show_devid(BMC,devrec,12);
   }

   ret = ipmi_getpicmg(devrec,sizeof(devrec),fdebug);
   if (ret == 0) {
        fpicmg = 1;
	printf("PICMG ATCA props = %02x %02x %02x %02x\n",
		devrec[0], devrec[1], devrec[2], devrec[3]);
   }
   
   ret = get_selftest_status(&selfbuf[0],sizeof(selfbuf));
   if (ret != 0) {
	printf("get_selftest_status error: ret = %x\n",ret);
	ipmi_close_();
	exit(1);
   } else {
        selfstatus = selfbuf[0] + (selfbuf[1] << 8);
        switch(selfstatus) {
           case 0x0055: selfstr = "(OK)";   break;
           default:     selfstr = "";   break;
        }
	printf("Selftest status  = %04x%s\n",selfstatus,selfstr);
   }

   {
      /* On some systems, this works locally but fails over IPMI LAN. */
      ret = get_chassis_status(&cbuf[0],sizeof(cbuf));
      if (ret != 0) printf("get_chassis_status error: ret = %d\n",ret);
      else show_chassis_status(cbuf,3);
   }

   ret = get_chan_auth(&devrec[0],sizeof(devrec));
   if (ret == 0) {  
      show_chan_auth("Channel Auth Cap",devrec,8);
   } else 
      printf("get_chan_auth error: ret = %x\n",ret);

   if (do_guid) {
      sresp = sizeof(devrec);
      ret = ipmi_cmd(GET_SYSTEM_GUID,NULL,0,devrec,&sresp,&cc,fdebug);
      if (fdebug) printf("system_guid: ret = %d, cc = %x\n",ret,cc);
      if (ret == 0 && cc == 0) {
         printf("System GUID      = ");
         for (i=0; i<16; i++) {
            if ((i == 4) || (i == 6) || (i == 8) || (i == 10)) s = "-";
            else s = "";
            printf("%s%02x",s,devrec[i]);
         }
         printf("\n");
      }
   }

   if (do_hsc) {
      /* Get HSC status */
      ret = get_hsc_devid(&devrec[0],sizeof(devrec));
      if (ret == 0) {  /* only if HSC is detected */
         show_devid(HSC,devrec,14);
      }
   }
   ipmi_close_();

   exit (ret);
}  /* end main()*/

/* end bmchealth.c */
