/*
 * getevent.c
 *
 * This utility waits for IPMI Event Messages.
 * Critical events are logged in the SEL, but certain other non-critical
 * events, such as some transitions from error to OK, can be received by 
 * requesting IPMI events.  This program is an example of how that can 
 * be done.  Some IPMI systems do not support this function though.
 *
 * Different IPMI drivers may have varying behavior.  For instance, the
 * OpenIPMI driver uses the IPMI GetMessage commands internally and 
 * does not allow client programs to use those commands.  It has its 
 * own custom mechanism, see getevent_mv().
 *
 * Author:  Andy Cress  arcress@users.sourceforge.net
 * Copyright (c) 2005-2006 Intel Corporation. 
 *
 * 02/11/05 Andy Cress - created
 * 05/18/05 Andy Cress - modified bmc_enable bits
 * 05/26/05 Andy Cress - added call to decode_sel_entry
 * 09/09/05 Andy Cress - added sensor_type filtering & return type.
 * 03/16/05 Andy Cress - added loop, and -o for frunOnce
 * 06/27/06 Andy Cress 1.1 - specific message for cc=0x80 (no data)
 * 07/18/06 Andy Cress 1.1 - added getevent_mv, etc.
 * 07/26/06 Andy Cress 1.1 - added msgout() routine for fflush
 * 08/08/06 Andy Cress 1.2 - added -s for SEL method
 * 08/08/06 Andy Cress 1.2 - added -s for SEL method
 * 08/22/06 Andy Cress 1.3 - direct IOs added with ipmiutil-1.7.5
 * 09/13/06 Andy Cress 1.4 - handle empty SEL (0xCB), 
 *                           call syncevent_sel after every new event.
 */
/*M*
Copyright (c) 2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#endif
#include <string.h>
#include "ipmicmd.h"
 
#define IMBPTIMEOUT  200   /*200 ms*/
#define IPMB_CHANNEL  0
#define ulong  unsigned long
#define uint   unsigned int
#define ushort unsigned short
#define uchar  unsigned char

extern void decode_sel_entry(uchar *evt, char *obuf);  /* see events.c */
/*
 * Global variables 
 */
char * progname  = "getevent";
char * progver   = "1.8";
char   fdebug    = 0;
char   fpicmg    = 0;
char   frunonce  = 0;
char   fAsync    = 0; 
int    timeout   = 120;   /* 120 seconds default timeout */
int    wait_interval = 1; /* 1 second between calls */
uchar  ipmi_maj  = 0;
uchar  ipmi_min  = 0;
uint   imb_handle = 0;
int drvtype = 0;  /* driver_type from ipmicmd.h: 1=Intel_imb, 3=MV_OpenIPMI */
static char fselevts = 0;
static ushort sel_recid = 0;
static uint   sel_time  = 0;
#define LAST_REC  0xFFFF
#ifdef WIN32
char idxfile[] = "evt.idx";
#else
char idxfile[] = "/usr/share/ipmiutil/evt.idx";
#endif

static int do_wait(int nsec)
{
   int rv = 0;
#ifdef WIN32
   Sleep(nsec * 1000);   /*Windows Sleep(millisec)*/
#else
   sleep(nsec);    /* Linux sleep(sec) */
#endif
   return(rv);
}

static int get_event_receiver(uchar *sa, uchar *lun)
{
	uchar rdata[30];
	int rlen;
	uchar ccode;
	int ret;

	rlen = 2;
#if 0
        ret = ipmi_cmdraw( 0x01,NETFN_SEVT,BMC_SA,PUBLIC_BUS,BMC_LUN,
			idata,0, rdata,&rlen,&ccode, fdebug);
#endif
        ret = ipmi_cmd(GET_EVENT_RECEIVER,NULL,0, rdata,&rlen,&ccode, 0);
	if (ret == 0 && ccode != 0) ret = ccode;
	if (ret == 0) {
	   *sa  = rdata[0];
	   *lun = rdata[1];
	}
	return(ret);
}  

static int set_bmc_enables(uchar enab)
{
	uchar idata[8];
	uchar rdata[30];
	int rlen;
	uchar ccode;
	int ret;

	idata[0] = enab;
	rlen = 1;
        ret = ipmi_cmdraw( 0x2E,NETFN_APP,BMC_SA,PUBLIC_BUS,BMC_LUN,
			idata,1, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;
	return(ret);
}

static int get_bmc_enables(uchar *enab)
{
	uchar rdata[30];
	int rlen;
	uchar ccode;
	int ret;

	rlen = 1;
        ret = ipmi_cmdraw( 0x2F,NETFN_APP,BMC_SA,PUBLIC_BUS,BMC_LUN,
			NULL,0, rdata,&rlen,&ccode, fdebug);
	if (ret == 0 && ccode != 0) ret = ccode;

	if (ret == 0) *enab  = rdata[0];
	return(ret);
}

/* 
 * msgout
 * wrapper for printf() to include fflush
 */
void msgout(char *pattn, ...)
{
    va_list arglist;

    va_start( arglist, pattn );
    vfprintf( stdout, pattn, arglist );
    va_end( arglist );
    fflush( stdout );
}

#ifdef TEST
/* The TEST flag enables the IMB Async Message method via get_imb_event,
 * but this is usually not used. */

/* The LANDesk library has the same function names as the imbapi.c */
#ifdef LINK_LANDESK
#define StartAsyncMesgPoll ia_StartAsyncMesgPoll
#define SendAsyncImbpRequest ia_SendAsyncImbpRequest
#define GetAsyncImbpMessage ia_GetAsyncImbpMessage
#define GetAsyncImbpMessage_Ex ia_GetAsyncImbpMessage_Ex
#define IsAsyncMessageAvailable ia_IsAsyncMessageAvailable
#define RegisterForImbAsyncMessageNotification ia_RegisterForImbAsyncMessageNotification
#define UnRegisterForImbAsyncMessageNotification ia_UnRegisterForImbAsyncMessageNotification
#endif  /*endif LINK_LANDESK*/

/*
 * get_imb_event
 * This only gets certain IMB events, like
 * OS requests (e.g. shutdown), and get_software_id 
 */
static int get_imb_event(uchar etype, int timeout, uchar *evt)
{
   	int ret = -1;
	int i;
	uchar buffer[512];
	ulong mlen;
	ulong asyseqnum;
        uchar sessionHandle = 0;
        uchar privilege = 0;

	/* clean out pre-existing async messages */
	while(1) {
	   mlen = sizeof(buffer);
	   if (GetAsyncImbpMessage(buffer,&mlen, IMBPTIMEOUT, &asyseqnum,
                                IPMB_CHANNEL) != 0) 
		break;
	}
	ret = RegisterForImbAsyncMessageNotification(&imb_handle);
	if (ret != 0) { 
		msgout("RegisterAsync error %d\n",ret);
		return(ret);
	}
        ret = StartAsyncMesgPoll();

	for (i = 0; (timeout == 0) || (i < timeout); i++)
	{
	   if (IsAsyncMessageAvailable(imb_handle) == 0) {
		mlen = sizeof(buffer);
		ret = GetAsyncImbpMessage_Ex (buffer, &mlen,
                                IMBPTIMEOUT, &asyseqnum, IPMB_CHANNEL,
                                 &sessionHandle, &privilege);
		/* get the message command */
                if (fdebug) msgout("got async msg, len=%d\n",mlen);
		/* copy the event to evt */
                if (mlen > 16) mlen = 16;
                memcpy(evt,buffer,mlen);
		break;
	   } else do_wait(1); /*wait 1 sec before retry*/
	}
	UnRegisterForImbAsyncMessageNotification (imb_handle,0);
	return(ret);
}
#endif   /*endif TEST*/

static int get_sel_entry(ushort recid, ushort *nextid, uchar *rec)
{
   uchar ibuf[6];
   uchar rbuf[32];
   int rlen;
   ushort id;
   uchar cc;
   int rv; 

   ibuf[0] = 0;
   ibuf[1] = 0;
   ibuf[2] = (recid & 0x00ff);
   ibuf[3] = (recid & 0xff00) >> 8;
   ibuf[4] = 0;
   ibuf[5] = 0xFF;
   rlen = sizeof(rbuf);
   rv = ipmi_cmd(GET_SEL_ENTRY, ibuf, 6, rbuf, &rlen, &cc, fdebug);
   if (rv == 0) {
       if (cc != 0) rv = cc;
       else {  /*success*/
          id = rbuf[0] + (rbuf[1] << 8);  /*next rec id*/
          memcpy(rec,&rbuf[2],16);
          *nextid = id;
          id = rbuf[2] + (rbuf[3] << 8);  /*curr rec id*/
       }
   }
   if (fdebug) printf("get_sel(%x) rv=%d cc=%x id=%x next=%x\n",
			recid,rv,cc,id,*nextid);
   return(rv);
}

static void startevent_sel(ushort *precid, uint *ptime) 
{
    FILE *fd;
    uchar rec[24];
    uint t = 0;
    ushort r = 0;
    ushort r2 = 0;
    int rv = -1;

    fd = fopen(idxfile,"r");
    if (fd != NULL) {
        // Read the file, get savtime & savid
        rv = fscanf(fd,"%x %x",&t,&r);
        fclose(fd);
        if (r == LAST_REC) r = 0;
    } else {  /* treat as first time */
        r = LAST_REC;
        rv = get_sel_entry(r,&r2,rec);
        if (rv == 0) {
            memcpy(&t,&rec[2],4);
            r = rec[0] + (rec[1] << 8); /*use current rec id*/
        } else r = 0;
    }
    if (fdebug) printf("start: recid=%x time=%x\n",r,t);
    *ptime  = t;
    *precid = r;
}

static void syncevent_sel(ushort recid, uint itime)
{
    FILE *fd;
    // Rewrite the saved time & record id
    if (fdebug) printf("sync: recid=%x time=%x\n",recid,itime);
    fd = fopen(idxfile,"w");
    if (fd != NULL) {
        fprintf(fd,"%x %x\n",itime,recid);
        fclose(fd);
    }
}

static int getevent_sel(uchar *rdata, int *rlen, uchar *ccode)
{
    uchar rec[24];
    int rv = 0;
    ushort newid;
    ushort nextid;
    
    /* get current last record */
    rv = get_sel_entry(sel_recid,&nextid,rec);
    if (rv == 0xCB && sel_recid == 0) {  /* SEL is empty */
        *ccode = rv;  /* save the real ccode */
        rv = 0x80;    /* this is ok, just keep waiting  */
    }
    if (rv == 0) {
       if ((nextid == LAST_REC) || (sel_recid == nextid)) { 
           *ccode = 0x80;  /*nothing new*/
       } else {
         sel_recid = nextid;  /* else get new one */
         rv = get_sel_entry(sel_recid,&nextid,rec);
         if (rv == 0) {  /* new event */ 
            newid = rec[0] + (rec[1] << 8);
            if (fdebug) printf("recid=%x newid=%x next=%x\n",
	   		      sel_recid,newid,nextid);
            memcpy(rdata,rec,16);
            *rlen = 16;
            *ccode = 0;
         }
       }
    }
    return(rv); 
}

static int get_event(uchar etype, int timeout, uchar *evt, uchar *stype)
{
   	int ret = 0;
   	uchar rdata[64];
   	int rlen;
	uchar ccode;
	int fretry;
	int i;

	for (i = 0; (timeout == 0) || (i < timeout); i++)
	{
	   rlen = sizeof(rdata);
	   fretry = 0;  ccode = 0;
           if (fselevts) {
             ret = getevent_sel(rdata,&rlen,&ccode);
           } else 
#ifndef WIN32
           if (drvtype == DRV_MV) { /* if MV OpenIPMI driver (Linux only) */
             /* use special MV API instead (see ipmimv.c) */
             ret = getevent_mv(rdata,&rlen,&ccode);
           } else
#endif
           ret = ipmi_cmd(READ_EVENT_MSGBUF,NULL,0,rdata,&rlen,&ccode,fdebug);
           /* IPMI 1.5 spec, section 18.8 says cc 0x80 means 
            * "data not available (queue/buffer empty)" */
           if (ret == 0 && ccode != 0) { ret = ccode; }
           if (ret == 0x80) { fretry = 1; do_wait(1);  /*wait 1 sec*/ 
           } else {
        	if (ret == 0) {
	    	    /* parse event types for a specified type */
		    /* etype param == 0xff is a mask, meaning get any event */
            	    if ((etype & rdata[11]) != 0) {
	       	        memcpy(evt,rdata,rlen);
			*stype = rdata[11];  /* return sensor type */
            	    } else {       /* keep looking */
			do_wait(wait_interval);
			continue; 
		    }
        	}
		/* if here, either got one, or need to return error */
		break;  
	    }
	}  /*end for loop*/
	return(ret);
}


void show_event(uchar *evt)
{
   int i;
   char obuf[132];

   msgout("event data: ");
   for (i=0; i<16; i++) msgout("%02x ",evt[i]);
   msgout("\n");

   decode_sel_entry(evt,obuf);
   msgout(obuf); 
}


#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
{
   int ret = 0;
   char c;
   uchar devrec[16];
   uchar event[16];
   uchar sa, lun;
   uchar enables;
   uchar fevmsgok;
   uchar sensor_type;
   int i;

   msgout("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"asot:T:V:EP:N:R:U:x?")) != EOF ) 
      switch(c) {
          case 's': fselevts = 1;   break;  /*use SEL events*/
          case 't': timeout = atoi(optarg); break;  /*timeout*/
          case 'x': fdebug = 1;     break;  /* debug messages */
          case 'a': fAsync = 1;     break;  /* imb async messages */
          case 'o': frunonce = 1;   break;  /* only run once for first event */
          case 'T':      /* auth type */
		i = atoi(optarg);  
                if (i > 0 && i <= 5) gauth_type = i;
                break;
          case 'V':      /* priv level */
		i = atoi(optarg);
                if (i > 0 && i <= 5) gpriv_level = i;
                break;
	  case 'N':      /* nodename */
		gnode = optarg;
		fipmi_lan = 1; 
		break;
	  case 'U':
		guser = strdup_(optarg);  /*remote username */
		if (!guser)  perror("strdup_");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		}
		break;
 	  case 'R': 
 	  case 'P': 
		gpswd = strdup_(optarg);  /*remote password */
		if (!gpswd) perror("strdup_");
		else {  /* Hide password from 'ps' */
	        	memset(optarg, ' ', strlen(optarg));	
		} 
 		break;
          case 'E':      /* get password from IPMI_PASSWORD environment var */
                gpswd = getenv("IPMI_PASSWORD");
                if (gpswd == NULL) perror("getenv(IPMI_PASSWORD)");
                else {
                   if (strlen(gpswd) > 16) gpswd[16] = 0;
                   if (fdebug) msgout("using IPMI_PASSWORD\n");
                }
                break;
	  default:
                printf("Usage: %s [-x -t sec -NPRUETV]\n", progname);
                printf(" where -t N    set timeout to N seconds\n"); 
                printf("       -s      use SEL events\n");
                printf("       -o      run Once for the first event\n");
                printf("       -x      show eXtra debug messages\n");
          printf("       -N node  Nodename of target system\n");
          printf("       -U user  Username for remote node\n");
          printf("       -P/-R pswd  Remote Password\n");
          printf("       -E   use password from Environment IPMI_PASSWORD\n");
          printf("       -T 1 use auth Type: 1=MD2, 2=MD5(default)\n");
          printf("       -V 2 use priVilege level: 2=user, 4=admin(default)\n");
                exit(1);
      }

   ret = ipmi_getdeviceid(devrec,16,fdebug);
   if (ret != 0) {
	msgout("Cannot do ipmi_getdeviceid, ret = %d\n",ret);
	ipmi_close_();
	exit(1);
   } else {
      ipmi_maj = devrec[4] & 0x0f;
      ipmi_min = devrec[4] >> 4;
      msgout("-- BMC version %x.%x, IPMI version %d.%d \n",
             devrec[2],  devrec[3], ipmi_maj, ipmi_min);
   }

   ret = ipmi_getpicmg(devrec,sizeof(devrec),fdebug);
   if (ret == 0) fpicmg = 1;

   /* get event receiver */
   ret = get_event_receiver(&sa ,&lun);
   if (ret != 0)
        msgout("event receiver error %d\n",ret);
   else msgout("event receiver sa = %02x lun = %02x\n",sa,lun);

   ret = get_bmc_enables(&enables);
   if (ret != 0) msgout("bmc enables error %d\n",ret);
   else {
        msgout("bmc enables = %02x\n",enables);
	if ((enables & 0x02) == 2) fevmsgok = 1;
	else fevmsgok = 0;
   }
   if (fevmsgok == 0) { 
      msgout("Event Message Buffers not enabled.\n");
      enables |= 0x0f;  /* 0x08=SEL, 0x02=EvtMsgBuf, rest is gravy */
      ret = set_bmc_enables(enables);
      if (ret != 0) {
	 msgout("set_bmc_enables error 0x%x\n",ret);
      }
      else msgout("set_bmc_enables success\n");
   }
   if (fselevts) startevent_sel(&sel_recid,&sel_time);

   drvtype = get_driver_type();
   if (fdebug) msgout("driver_type = %d\n",drvtype);
#ifdef TEST
   if (fAsync && drvtype == DRV_IMB) {  /*use imb async messages*/
     msgout("Waiting for an imb event ...\n");  /* no timeout */
     ret = get_imb_event(0xff,timeout,event);
   } else
#endif
      /* loop on events here, like a daemon would. */
      while (ret == 0) {  /*wait for bmc message events*/
         msgout("Waiting %d seconds for an event ...\n",timeout);
         /* note: could also get message flags here */
         /* 0xff is a mask, meaning get any events */
         ret = get_event(0xff,timeout,event,&sensor_type);
         if (fdebug) msgout("get_event ret = %d\n",ret);
         if (ret == 0) { /* got an event successfully */
            msgout("got event, sensor_type = %02x\n",sensor_type);
	    show_event(event);
            if (fselevts) syncevent_sel(sel_recid,sel_time);
         } else {
	    msgout("get_event error: ret = 0x%x\n",ret);
         }
         if (frunonce) break;
         if (timeout == 0 && ret == 0x80) {  /*0x80 = no data yet */
           if (fdebug) msgout("get_event timeout, no data yet.\n");
           do_wait(1);
           ret = 0;  /*ok, keep going*/
         }
   }  /*end while loop*/
   
   if (fselevts) syncevent_sel(sel_recid,sel_time);
   ipmi_close_();
   exit(ret);
}  /* end main()*/

/* end getevent.c */
